/// <reference types="node" />
import { ChildProcess } from 'child_process';
import { Disposable } from '@devpodio/core';
import { Repository, WorkingDirectoryStatus, Branch, GitResult, GitError, GitFileStatus, GitFileChange, CommitWithChanges, GitFileBlame, Remote as RemoteModel, StashEntry } from './git-model';
export declare const GitPath = "/services/git";
export declare const Git: unique symbol;
export declare namespace Git {
    namespace Options {
        namespace BranchCommand {
            interface List {
                readonly type: 'current' | 'local' | 'remote' | 'all';
            }
            interface Create {
                readonly toCreate: string;
                readonly startPoint?: string;
            }
            interface Delete {
                readonly toDelete: string;
                readonly force?: boolean;
                readonly remote?: boolean;
            }
            interface Rename {
                readonly newName: string;
                readonly oldName?: string;
                readonly force?: boolean;
            }
        }
        interface Clone {
            readonly localUri: string;
            readonly branch?: string;
        }
        interface Repositories {
            readonly maxCount?: number;
        }
        interface Unstage {
            readonly reset?: 'index' | 'working-tree' | 'all';
            readonly treeish?: string;
        }
        namespace Checkout {
            interface CheckoutBranch {
                readonly branch: string;
                readonly force?: boolean;
                readonly merge?: boolean;
                readonly newBranch?: string;
            }
            interface WorkingTreeFile {
                readonly paths: string | string[];
                readonly force?: boolean;
                readonly merge?: boolean;
                readonly treeish?: string;
            }
        }
        interface Commit {
            readonly amend?: boolean;
            readonly signOff?: boolean;
        }
        interface Show {
            readonly encoding?: 'utf8' | 'binary';
            readonly commitish?: 'index' | string;
        }
        interface Stash {
            readonly action?: 'push' | 'apply' | 'pop' | 'list' | 'drop' | 'clear';
            readonly id?: string;
            readonly message?: string;
        }
        interface Fetch {
            readonly remote?: string;
        }
        interface Push {
            readonly remote?: string;
            readonly localBranch?: string;
            readonly remoteBranch?: string;
            readonly setUpstream?: boolean;
            readonly force?: boolean;
        }
        interface Pull {
            readonly remote?: string;
            readonly branch?: string;
            readonly rebase?: boolean;
        }
        interface Reset {
            readonly mode: 'hard' | 'soft' | 'mixed';
            readonly ref?: string;
        }
        interface Merge {
            readonly branch: string;
        }
        interface Execution {
            readonly successExitCodes?: ReadonlySet<number>;
            readonly expectedErrors?: ReadonlySet<GitError>;
            readonly env?: Object;
            readonly stdin?: string;
            readonly stdinEncoding?: string;
            readonly maxBuffer?: number;
            readonly processCallback?: (process: ChildProcess) => void;
            readonly name?: string;
        }
        interface Range {
            readonly toRevision?: string;
            readonly fromRevision?: number | string;
        }
        interface Diff {
            readonly range?: Range;
            readonly uri?: string;
        }
        interface Log extends Diff {
            readonly branch?: string;
            readonly maxCount?: number;
            readonly shortSha?: boolean;
        }
        interface Blame {
            readonly content?: string;
        }
        interface LsFiles {
            readonly errorUnmatch?: true;
        }
        interface Remote {
            readonly verbose?: true;
        }
    }
}
export interface Git extends Disposable {
    clone(remoteUrl: string, options: Git.Options.Clone): Promise<Repository>;
    repositories(workspaceRootUri: string, options: Git.Options.Repositories): Promise<Repository[]>;
    status(repository: Repository): Promise<WorkingDirectoryStatus>;
    add(repository: Repository, uri: string | string[]): Promise<void>;
    unstage(repository: Repository, uri: string | string[], options?: Git.Options.Unstage): Promise<void>;
    branch(repository: Repository, options: {
        type: 'current';
    }): Promise<Branch | undefined>;
    branch(repository: Repository, options: {
        type: 'local' | 'remote' | 'all';
    }): Promise<Branch[]>;
    branch(repository: Repository, options: Git.Options.BranchCommand.Create | Git.Options.BranchCommand.Rename | Git.Options.BranchCommand.Delete): Promise<void>;
    checkout(repository: Repository, options: Git.Options.Checkout.CheckoutBranch | Git.Options.Checkout.WorkingTreeFile): Promise<void>;
    commit(repository: Repository, message?: string, options?: Git.Options.Commit): Promise<void>;
    fetch(repository: Repository, options?: Git.Options.Fetch): Promise<void>;
    push(repository: Repository, options?: Git.Options.Push): Promise<void>;
    pull(repository: Repository, options?: Git.Options.Pull): Promise<void>;
    reset(repository: Repository, options: Git.Options.Reset): Promise<void>;
    merge(repository: Repository, options: Git.Options.Merge): Promise<void>;
    show(repository: Repository, uri: string, options?: Git.Options.Show): Promise<string>;
    stash(repository: Repository, options?: Readonly<{
        action?: 'push';
        message?: string;
    }>): Promise<void>;
    stash(repository: Repository, options: Readonly<{
        action: 'list';
    }>): Promise<StashEntry[]>;
    stash(repository: Repository, options: Readonly<{
        action: 'clear';
    }>): Promise<void>;
    stash(repository: Repository, options: Readonly<{
        action: 'apply' | 'pop' | 'drop';
        id?: string;
    }>): Promise<void>;
    remote(repository: Repository): Promise<string[]>;
    remote(repository: Repository, options: {
        verbose: true;
    }): Promise<RemoteModel[]>;
    exec(repository: Repository, args: string[], options?: Git.Options.Execution): Promise<GitResult>;
    diff(repository: Repository, options?: Git.Options.Diff): Promise<GitFileChange[]>;
    log(repository: Repository, options?: Git.Options.Log): Promise<CommitWithChanges[]>;
    blame(repository: Repository, uri: string, options?: Git.Options.Blame): Promise<GitFileBlame | undefined>;
    lsFiles(repository: Repository, uri: string, options: {
        errorUnmatch: true;
    }): Promise<boolean>;
    lsFiles(repository: Repository, uri: string, options?: Git.Options.LsFiles): Promise<any>;
}
export declare namespace GitUtils {
    function isBranchRename(arg: any | undefined): arg is Git.Options.BranchCommand.Rename;
    function isBranchDelete(arg: any | undefined): arg is Git.Options.BranchCommand.Delete;
    function isBranchCreate(arg: any | undefined): arg is Git.Options.BranchCommand.Create;
    function isBranchList(arg: any | undefined): arg is Git.Options.BranchCommand.List;
    function isBranchCheckout(arg: any | undefined): arg is Git.Options.Checkout.CheckoutBranch;
    function isWorkingTreeFileCheckout(arg: any | undefined): arg is Git.Options.Checkout.WorkingTreeFile;
    function isRepositoryDoesNotExistError(error: any | undefined): boolean;
    function mapStatus(rawStatus: string): GitFileStatus;
    function isSimilarityStatus(rawStatus: string): boolean;
}
